infilename_mask = '*.afa'
input_directory = 'aln'
table_file_name = 'repeats.txt'
stats_file_name = 'len_stats.txt'

import os
import re
from glob import glob

def main():
    try:
        subdirs = sorted(os.walk(input_directory).next()[1])
    except StopIteration:
        exit("No subdirectories found in '{0}'.".format(input_directory))
    stats = {}
    with file(table_file_name, 'w') as outfile:
        for subdir in subdirs:
            print subdir
            stats[subdir] = {}
            full_mask = os.path.join(input_directory, subdir, infilename_mask)
            for file_name in glob(full_mask):
                short_file_name = os.path.split(file_name)[-1]
                repeats = find_repeats(file_name)
                for start, end, root, mid_count, side_count in repeats:
                    rep_len = len(root)
                    count = stats[subdir].get(rep_len, 0)
                    stats[subdir][rep_len] = count + 1
                    outfile.write(subdir + '\t' + short_file_name + '\t' +
                                  "\t".join(map(str, (start, end, rep_len,
                                                      mid_count, side_count,
                                                      root))) + '\n')
    with file(stats_file_name, 'w') as outfile:
        for family, fstats in sorted(stats.items()):
            outfile.write(family + '\n')
            for rep_len, count in sorted(fstats.items()):
                outfile.write("\t{0}\t{1}".format(rep_len, count) + '\n')
            outfile.write('\n')


def find_repeats(infile_name, debugging=False):
    repeats = set()
    alignments = read_alignments(infile_name)
    sequences = alignments.values()
    for i in range(len(sequences)):
        for j in range(i + 1, len(sequences)):
            reps = ideal_repeats((sequences[i], sequences[j]), debugging)
            repeats |= set(reps)
    return sorted(repeats)

def ideal_repeats(sequences, debugging=False):
    columns = zip(*sequences)
    conserved = ''.join(map(column_conserved, columns))
    conserved_nodel = ''.join(map(column_conserved_nodel, columns))

    ir_re = re.compile("(?<=[ACTG])-+(?=[ACTG])", re.I)
    # find all insertions with ACP on both sides
    # ACP = absolutely conservative position, NACP = non-ACP
    insertions = re_findall_coordinates(conserved, ir_re)
    # filter out insertions with NACP inside
    insertions = [(start, end) for start, end in insertions
                  if " " not in conserved_nodel[start:end]]
    ## filter out insertions having a period less then 3
    #insertions = [(start, end) for start, end in insertions
    #              if seq_period(conserved_nodel[start:end]) >= 3]

    if debugging:
        def ranges_debug_repr(ranges, length):
            result = list(' '*length)
            for start, stop in ranges:
                if stop == start + 1:
                    result[start] = 'X'
                else:
                    result[start], result[stop - 1] = '[', ']'
            return ''.join(result)

        outfile_name = infile_name + '.txt'
        with file(outfile_name, 'w') as outfile:
            for hdr, seq in sorted(alignments.items()):
                outfile.write(seq + '\n')
            outfile.write('\n')
            outfile.write(conserved + '\n')
            outfile.write(conserved_nodel + '\n')
            test_insertions = ranges_debug_repr(insertions, len(columns))
            outfile.write(test_insertions + '\n')

    result = []
    for start, end in insertions:
        length = end - start
        insertion = conserved_nodel[start:end]
        period = seq_period(insertion)
        root = insertion[:period]
        left = count_repeats_left(conserved, start, root)
        right = count_repeats_right(conserved, end, root)
        if left + right:
            result.append((start, end, root, length / period, left + right))
    return result

def read_alignments(infile_name):
    with file(infile_name) as infile:
        lines = infile.readlines()
    header = None
    sequences = {}
    for ln, line in enumerate(lines, 1):
        line = line.strip()
        if not line or line == '#mega' or line.startswith('!'):
            continue
        if line.startswith('>'):
            header = line[1:]
            if header in sequences:
                exit('Duplicate header {0} on line {1}.'.format(header, ln))
            sequences[header] = ''
        elif header:
            sequences[header] += line
    if min(len(seq) for seq in sequences.values()) != \
       max(len(seq) for seq in sequences.values()):
        exit('Sequences have unequal length.')
    return sequences

def column_conserved(column):
    letters = set(column)
    if len(letters) == 1:
        return column[0]
    elif '-' in letters:
        return '-'
    else:
        return ' '

def column_conserved_nodel(column):
    letters = set(column)
    if '-' in letters:
        letters.remove('-')
    if len(letters) == 1:
        return list(letters)[0]
    else:
        return ' '

def re_findall_coordinates(sequence, regexp):
    insertions = []
    pos = 0
    while True:
        match = regexp.search(sequence, pos)
        if not match:
            break
        start, end = match.start(), match.end()
        insertions.append((start, end))
        pos = end
    return insertions

def seq_period(seq):
    """Return min(p): seq == seq[:p] * (len(seq) / p)"""
    if not seq:
        return 0
    l = len(seq)
    for p in range(1, l / 2 + 1):
        if l % p:
            continue
        #if seq == seq[:p] * (l / p):
        root = seq[:p]
        f = True
        for i in range(1, l / p):
            if seq[i*p:(i+1)*p] != root:
                f = False
                continue
        if f:
            return p
    return l

def count_repeats_left(seq, end, sub):
    count = 0
    l = len(sub)
    while True:
        if seq[end-(count+1)*l:end-count*l] != sub:
            break
        count += 1
    return count

def count_repeats_right(seq, start, sub):
    count = 0
    l = len(sub)
    while True:
        if seq[start+count*l:start+(count+1)*l] != sub:
            break
        count += 1
    return count


if __name__ == '__main__':
    main()
